#!/usr/bin/env node
"use strict";
/**
 * lock-internal-packages
 * Add internal packages to lockfiles
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = __importStar(require("path"));
const t = __importStar(require("io-ts"));
const A = __importStar(require("fp-ts/Array"));
const E = __importStar(require("fp-ts/Either"));
const NEA = __importStar(require("fp-ts/NonEmptyArray"));
const F = __importStar(require("fluture"));
const D = __importStar(require("io-ts-docopt"));
const io_ts_docopt_1 = require("io-ts-docopt");
const PathReporter_1 = require("io-ts/lib/PathReporter");
const function_1 = require("fp-ts/function");
const nonEmptyArray_1 = require("io-ts-types/lib/nonEmptyArray");
const lerna_utils_1 = require("@typescript-tools/lerna-utils");
const stringify_json_1 = require("@typescript-tools/stringify-json");
const dependency_graph_1 = require("@typescript-tools/dependency-graph");
const lerna_packages_1 = require("@typescript-tools/lerna-packages");
const PackageLockJson_1 = require("@typescript-tools/io-ts/dist/lib/PackageLockJson");
const StringifiedJSON_1 = require("@typescript-tools/io-ts/dist/lib/StringifiedJSON");
const DistTag_1 = require("@typescript-tools/io-ts/dist/lib/DistTag");
const NpmRegistry_1 = require("@typescript-tools/io-ts/dist/lib/NpmRegistry");
const RegistryMetadata_1 = require("@typescript-tools/io-ts/dist/lib/RegistryMetadata");
const execa_1 = __importDefault(require("execa"));
// REFACTOR: this entire data-flow is pretty shocking. It's clear I
// need to find some patterns with Futures analogous to sequenceS with
// fp-ts monads.
const docstring = `
Usage:
    lock-internal-packages [--root=<path>] [--dist-tag=<tag>] --registry=<url>

Options:
    --root=<path>       Path to root of monorepo [default: .]
    --dist-tag=<tag>    NPM dist-tag to use for published internal-packages [default: latest]
    --registry=<url>    NPM registry hosting internal packages
`;
// DISCUSS
// TODO: take as input the base url of the npm registry
const CommandLineOptions = io_ts_docopt_1.withEncode(t.type({
    '--root': t.string,
    '--dist-tag': DistTag_1.DistTag,
    '--registry': NpmRegistry_1.NpmRegistry,
}), a => ({
    root: a['--root'],
    tag: a['--dist-tag'],
    registry: a['--registry'],
}));
// Temporary helper to infer types, see https://github.com/fluture-js/Fluture/issues/455
function map(mapper) {
    return F.map(mapper);
}
// Widens the type of a particular Err into an Err
const error = (error) => error;
const decodeDocopt = (codec, docstring) => function_1.pipe(D.decodeDocopt(codec, docstring), E.mapLeft(err => error({ type: 'docopt decode', err: PathReporter_1.PathReporter.report(E.left(err)).join('\n') })), E.map(options => F.resolve(options)), E.getOrElse(err => F.reject(err)));
const dependencyGraph = (root) => function_1.pipe(dependency_graph_1.dependencyGraph(root), F.mapRej(err => error({ type: 'unable to create dependency graph', err })));
const lernaPackages = (root) => function_1.pipe(lerna_packages_1.lernaPackages(root), map(A.reduce(new Map(), (acc, pkg) => (acc.set(pkg.name, pkg), acc))));
const decodeRegistryMetadata = (value) => function_1.pipe(StringifiedJSON_1.StringifiedJSON(nonEmptyArray_1.nonEmptyArray(RegistryMetadata_1.RegistryMetadata)).decode(value), E.map(NEA.head), E.mapLeft(errors => PathReporter_1.PathReporter.report(E.left(errors)).join('\n')), E.mapLeft(err => error({ type: 'unable to decode registry metadata', err })));
const registryMetadata = (tag) => (packageName) => F.Future((reject, resolve) => {
    const subcommand = execa_1.default.command(`npm pack ${packageName}@${tag} --dry-run --json`);
    subcommand
        .then(({ stdout }) => function_1.pipe(decodeRegistryMetadata(stdout), E.fold(reject, resolve)));
    return function onCancel() {
        subcommand.cancel();
    };
});
const readFile = (file) => function_1.pipe(lerna_utils_1.readFile(file), F.mapRej(err => error({ type: 'unable to read file', file, err })));
const decodePackageLockJson = (file) => (contents) => function_1.pipe(StringifiedJSON_1.StringifiedJSON(PackageLockJson_1.PackageLockJson).decode(contents), E.mapLeft(errors => PathReporter_1.PathReporter.report(E.left(errors)).join('\n')), E.mapLeft(err => error({ type: 'unable to parse lockfile', file, err })), E.map(F.resolve), E.getOrElseW(F.reject));
const readPackageLockJson = (name) => (file) => function_1.pipe(readFile(file), F.chain(decodePackageLockJson(file)), map(lockfile => ({ name, lockfile })));
const stringifyJSON = (onError) => (json) => function_1.pipe(stringify_json_1.stringifyJSON(onError)(json), E.map(F.resolve), E.getOrElseW(err => F.reject(error({ type: 'unable to stringify json', json, err }))));
const writeJsonFile = (file) => (contents) => function_1.pipe(stringifyJSON(E.toError)(contents), F.chain(contents => function_1.pipe(lerna_utils_1.writeFile(file)(contents), F.mapRej(err => error({ type: 'unable to write file', file, err })))));
// REFACTOR: these `parallel` calls should be run in parallel, not serially.
// DISCUSS: is this a time for TaskEither to shine?
const lockInternalPackages = (tag) => (registry) => (packageMetadata) => (dependencies) => (lockfile) => function_1.pipe(Array.from(dependencies.keys()), A.map(registryMetadata(tag)), F.parallel(Infinity), map(A.reduce(new Map(), (acc, metadata) => (acc.set(metadata.name, metadata), acc))), F.chain(dependencyMetadata => function_1.pipe(Array.from(dependencies.values()), A.map(manifest => function_1.pipe(path.resolve(manifest.location, 'package-lock.json'), readPackageLockJson(manifest.name))), F.parallel(Infinity), map(A.reduce(new Map(), (acc, { name, lockfile }) => (acc.set(name, lockfile), acc))), map(dependencyLockfiles => ({ dependencyMetadata, dependencyLockfiles })))), 
// combine the maps :grimacing:
map(({ dependencyMetadata, dependencyLockfiles }) => function_1.pipe(Array.from(dependencyMetadata.keys()), A.reduce(new Map(), (acc, name) => (acc.set(name, {
    registry: dependencyMetadata.get(name),
    lockfile: dependencyLockfiles.get(name),
    manifest: dependencies.get(name),
}),
    acc)))), F.chain(dependencyMetadata => {
    for (const [dependency, depMetadata] of dependencyMetadata.entries()) {
        lockfile.dependencies[dependency] = {
            version: depMetadata.registry.version,
            resolved: `${registry}/${depMetadata.manifest.name}/-/${depMetadata.registry.filename}`,
            integrity: depMetadata.registry.integrity,
            requires: Object.assign({}, depMetadata.manifest.dependencies, depMetadata.manifest.peerDependencies, depMetadata.manifest.optionalDependencies),
            dependencies: depMetadata.lockfile.dependencies,
        };
    }
    return writeJsonFile(path.resolve(packageMetadata.location, 'package-lock.json'))(lockfile);
}));
function main() {
    function_1.pipe(decodeDocopt(CommandLineOptions, docstring), F.chain(options => function_1.pipe(lernaPackages(options.root), F.chain(lernaPackages => function_1.pipe(dependencyGraph(options.root), map(dependencyGraph => ({ lernaPackages, dependencyGraph })))), value => value, F.chain(({ lernaPackages, dependencyGraph }) => function_1.pipe(Array.from(lernaPackages.values()), A.map(pkg => function_1.pipe(readPackageLockJson(pkg.name)(path.resolve(pkg.location, 'package-lock.json')), map(value => value.lockfile), F.chain(lockInternalPackages(options.tag)(options.registry)(lernaPackages.get(pkg.name))(function_1.pipe(Array.from(dependencyGraph.get(pkg.name).values()), A.reduce(new Map(), (acc, manifest) => (acc.set(manifest.name, manifest), acc))))))), F.parallel(Infinity))))), F.fork(error => (console.error(error), process.exit(1)))(function_1.constVoid));
}
main();
// Local Variables:
// mode: typescript
// End:
//# sourceMappingURL=index.js.map